#  Copyright (c) 2021 zfit
import logging
from functools import lru_cache, wraps
from typing import Callable, Dict, List, Optional, Tuple, Union

import numdifftools
import numpy as np
import scipy.stats
import tensorflow as tf
from scipy import optimize

from .. import settings, z
from ..core.interfaces import ZfitIndependentParameter
from ..core.parameter import assign_values
from ..param import set_values
from ..util.container import convert_to_container
from ..util.deprecation import deprecated_args


class NewMinimum(Exception):
    """Exception class for cases where a new minimum is found."""
    pass


class FailEvalLossNaN(Exception):
    pass


@deprecated_args(None, "Use cl for confidence level instead.", 'sigma')
def compute_errors(result: "zfit.result.FitResult",
                   params: List[ZfitIndependentParameter],
                   cl: Optional[float] = None,
                   rtol: Optional[float] = 0.001,
                   method: Optional[str] = None,
                   covariance_method: Optional[Union[str, Callable]] = None,
                   sigma: float = 1,
                   ) -> Tuple[Dict[ZfitIndependentParameter, Dict[str, float]],
                              Union["zfit.result.FitResult", None]]:
    """Compute asymmetric errors of parameters by profiling the loss function in the fit result.

    This method finds the value for a given parameter where the loss function is `cl` away: for example
    for a cl of 68.3%, this is one (multiplied by the errordef). The other parameters are also minimized and
    not fixed. This method is comparably computationally intensive and, if possible, `hesse` should be used.
    However, since `hesse` does not capture asymetric or non-parabolic shaped profiles well, this method is
    preferable.

    Args:
        result: fit result to be used to compute the uncertainties.
        params: The parameters to calculate the
            errors error. If None, use all parameters.
        cl: Confidence Level of the parameter to be determined. Defaults to 68.3%.
        rtol: relative tol between the computed and the exact roots
        method: type of solver, `method` argument of :py:func:`scipy.optimize.root`. Defaults to "hybr".
        covariance_method: The method to use to calculate the correlation matrix, will be forwarded directly
            to :py:meth:`FitResult.covariance`. Valid choices are
            by default {'minuit_hesse', 'hesse_np'} (or any other method defined in the result)
            or a Callable.
        sigma: Errors are calculated with respect to `sigma` std deviations.


    Returns:
        out:
            A `dict` containing as keys the parameter and as value a `dict` which
            contains two keys 'lower' and 'upper', holding the calculated errors.
            Example: result[par1]['upper'] -> the asymmetric upper error of 'par1'
        out: a fit result is returned when a new minimum is found during the loss scan
    """
    method = "hybr" if method is None else method
    # method = "krylov" if method is None else method  # TODO: integration tests, better for large n params?
    if cl is None:
        factor = 1.0
    else:
        factor = scipy.stats.chi2(1).ppf(cl)
    params = convert_to_container(params)
    new_result = None

    all_params = list(result.params.keys())
    loss = result.loss
    errordef = loss.errordef * factor
    fmin = result.fmin
    rtol *= errordef
    minimizer = result.minimizer
    from zfit import run
    old_values = run(result)

    covariance = result.covariance(method=covariance_method, as_dict=True)
    param_errors = {param: covariance[(param, param)] ** 0.5 for param in params}
    # param_scale = np.array(list(param_errors.values()))  # TODO: can be used for root finding initialization?

    ncalls = 0
    try:
        # start = time.time()
        to_return = {}
        for param in params:
            assign_values(all_params, result)

            logging.info(f"profiling the parameter {param}")
            param_error = param_errors[param]
            param_value = result.params[param]["value"]
            other_params = [p for p in all_params if p != param]

            initial_values = {"lower": [], "upper": []}
            direction = {"lower": -sigma, "upper": sigma}

            for ap in all_params:
                ap_value = result.params[ap]["value"]
                error_factor = covariance[(param, ap)] * (2 * errordef / param_error ** 2) ** 0.5
                for d in ["lower", "upper"]:
                    ap_value_init = ap_value + direction[d] * error_factor
                    initial_values[d].append(ap_value_init)

            # TODO: improvement, use jacobian?
            # @np_cache(maxsize=25)
            def func(values, args):
                nonlocal ncalls
                ncalls += 1
                swap_sign = args

                assign_values(all_params, values)
                try:
                    loss_value, gradient = loss.value_gradient(params=other_params)
                except tf.errors.InvalidArgumentError:
                    msg = (f"The evaluation of the errors of {param.name} failed due to too many NaNs"
                           " being produced in the loss and/or its gradient. This is most probably"
                           " caused by negative values returned from the PDF.")
                    raise FailEvalLossNaN(msg)

                zeroed_loss = loss_value.numpy() - fmin

                gradient = np.array(gradient)
                if swap_sign(param):  # mirror at x-axis to remove second zero
                    zeroed_loss = - zeroed_loss
                    gradient = - gradient
                    logging.info("Swapping sign in error calculation 'zfit_error'")

                elif zeroed_loss < - minimizer.tol:
                    assign_values(all_params, values)  # set values to the new minimum
                    raise NewMinimum("A new minimum is found.")

                downward_shift = errordef * sigma ** 2
                shifted_loss = zeroed_loss - downward_shift

                return np.concatenate([[shifted_loss], gradient])

            to_return[param] = {}
            swap_sign = {
                "lower": lambda p: p > param_value,
                "upper": lambda p: p < param_value,
            }
            for d in ["lower", "upper"]:
                roots = optimize.root(fun=func,
                                      args=(swap_sign[d],),
                                      x0=np.array(initial_values[d]),
                                      tol=rtol,
                                      options={
                                          'factor': 0.1,
                                          # 'diag': 1 / param_scale,  # scale factor for variables
                                          # 'diag': param_scale,
                                      },
                                      method=method)
                to_return[param][d] = roots.x[all_params.index(param)] - param_value
                # print(f"error {d}, time needed {time.time() - start2}")
        # print(f"errors found, time needed {time.time() - start}")
        assign_values(all_params, old_values)

    except NewMinimum as e:
        from .. import settings
        if settings.get_verbosity() >= 5:
            print(e)
        minimizer = result.minimizer
        loss = result.loss
        new_found_fmin = loss.value()
        new_result = minimizer.minimize(loss=loss)
        if new_result.fmin >= new_found_fmin:
            raise RuntimeError("A new minimum was discovered but the minimizer was not able to find this on himself. "
                               "This behavior is currently an exception but will most likely change in the future.")
        to_return, new_result_ = compute_errors(result=new_result, params=params, sigma=sigma, rtol=rtol,
                                                method=method)
        if new_result_ is not None:
            new_result = new_result_
    return to_return, new_result


def numerical_pdf_jacobian(func, params):
    jacobian_func = numdifftools.Jacobian(func)
    return jacobian_func([param.value() for param in params]).T


@z.function(wraps='autodiff')
def autodiff_pdf_jacobian(func, params):
    columns = []

    for p in params:
        vector = np.zeros(len(params))
        vector[params.index(p)] = 1.
        with tf.autodiff.ForwardAccumulator(params, list(vector)) as acc:
            values = func()
        columns.append(acc.jvp(values))

    jacobian = z.convert_to_tensor(columns)

    return jacobian


def covariance_with_weights(method, result, params):
    model = result.loss.model
    data = result.loss.data
    from zfit import run
    old_vals = run(params)

    Hinv_dict = method(result=result, params=params)  # inverse of the hessian matrix
    if not Hinv_dict:
        return {}
    Hinv = dict_to_matrix(params, Hinv_dict)

    def func():
        values = []
        for m, d in zip(model, data):
            v = m.log_pdf(d)
            if d.weights is not None:
                v *= d.weights
            values.append(v)
        return tf.concat(values, axis=0)

    if settings.options['numerical_grad']:
        def wrapped_func(values):
            assign_values(params, values)
            return func()

        jacobian = numerical_pdf_jacobian(func=wrapped_func, params=params)
    else:
        jacobian = autodiff_pdf_jacobian(func=func, params=params).numpy()

    C = np.matmul(jacobian, jacobian.T)
    covariance = np.matmul(Hinv, np.matmul(C, Hinv))
    assign_values(params, old_vals)
    return matrix_to_dict(params, covariance)


def dict_to_matrix(params, matrix_dict):
    nparams = len(params)
    matrix = np.empty((nparams, nparams))
    if not matrix_dict:
        return None

    for i in range(nparams):
        pi = params[i]
        for j in range(i, nparams):
            pj = params[j]
            matrix[i, j] = matrix_dict[(pi, pj)]
            if i != j:
                matrix[j, i] = matrix_dict[(pi, pj)]

    return matrix


def matrix_to_dict(params, matrix):
    nparams = len(params)
    matrix_dict = {}
    if matrix is None:
        return matrix_dict

    for i in range(nparams):
        pi = params[i]
        for j in range(i, nparams):
            pj = params[j]
            matrix_dict[(pi, pj)] = matrix[i, j]
            if i != j:
                matrix_dict[(pj, pi)] = matrix[i, j]

    return matrix_dict


def np_cache(*args, **kwargs):
    """LRU cache implementation for functions whose FIRST parameter is a numpy array.

    >>> array = np.array([[1, 2, 3], [4, 5, 6]])
    >>> @np_cache(maxsize=256)
    ... def multiply(array, factor):
    ...     print("Calculating...")
    ...     return factor*array
    >>> multiply(array, 2)
    Calculating...
    array([[ 2,  4,  6],
           [ 8, 10, 12]])
    >>> multiply(array, 2)
    array([[ 2,  4,  6],
           [ 8, 10, 12]])
    >>> multiply.cache_info()
    CacheInfo(hits=1, misses=1, maxsize=256, currsize=1)
    """

    def decorator(function):
        @wraps(function)
        def wrapper(np_array, *args, **kwargs):
            hashable_array = array_to_tuple(np_array)
            return cached_wrapper(hashable_array, *args, **kwargs)

        @lru_cache(*args, **kwargs)
        def cached_wrapper(hashable_array, *args, **kwargs):
            array = np.array(hashable_array)
            return function(array, *args, **kwargs)

        def array_to_tuple(np_array):
            """Iterates recursively."""
            try:
                return tuple(array_to_tuple(_) for _ in np_array)
            except TypeError:
                return np_array

        # copy lru_cache attributes over too
        wrapper.cache_info = cached_wrapper.cache_info
        wrapper.cache_clear = cached_wrapper.cache_clear

        return wrapper

    return decorator



"""z is a zfit TensorFlow version, that wraps TF while adding some conveniences, basically using a different default
dtype (`zfit.ztypes`). In addition, it expands TensorFlow by adding a few convenient functions helping to deal with
`NaN`s and similar.

Some function are already wrapped, others are not. Best practice is to use `z` whenever possible and `tf` for the rest.
"""

#  Copyright (c) 2021 zfit

# fill the following in to the namespace for (future) wrapping

# doesn't work below because of autoimport... probably anytime in the Future :)
# from tensorflow import *  # Yes, this is wanted. Yields an equivalent z BUT we COULD wrap it :)
# _module_dict = tensorflow.__dict__
# try:
#     _to_import = tensorflow.__all__
# except AttributeError:
#     _to_import = [name for name in _module_dict if not name.startswith('_')]
#
# __all__ = _to_import
# del tensorflow
# _imported = {}
# _failed_imports = []
# for _name in _to_import:
#     try:
#         _imported[_name] = _module_dict[_name]
#     except KeyError as error:
#         _failed_imports.append(_name)
# if _failed_imports:
#     warnings.warn("The following modules/attributes from TensorFlow could NOT be imported:\n{}".format(
#     _failed_imports))
# globals().update(_imported)
#
# del _imported, _failed_imports, _to_import, _module_dict

# same as in TensorFlow, wrapped


# noinspection PyUnresolvedReferences
import tensorflow.experimental.numpy as numpy

from . import math, random, unstable
from .wrapping_tf import (check_numerics, complex, convert_to_tensor, exp, pow,
                          random_normal, random_uniform, reduce_prod,
                          reduce_sum, sqrt, square)
from .zextension import abs_square, constant
from .zextension import function as function
from .zextension import (function_sampling, function_tf_input, nth_pow, pi,
                         py_function, run_no_nan, safe_where, stack_x,
                         to_complex, to_real, unstack_x)



"""The setup script."""

#  Copyright (c) 2021 zfit
import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as requirements_file:
    requirements = requirements_file.read().splitlines()

with open(os.path.join(here, 'requirements_dev.txt'), encoding='utf-8') as requirements_dev_file:
    requirements_dev = requirements_dev_file.read().splitlines()

tests_require = [
    'pytest>=3.4.2,<5.4',  # breaks unittests
    'pytest-runner>=2.11.1',
    'pytest-rerunfailures>=6',
    'pytest-xdist',
    'pytest-ordering',
    'pytest-randomly',
    'scikit-hep-testdata',
    'pytest-timeout>=1',
    'matplotlib'  # for plots in examples
]
setup(
    author=("Jonas Eschle <Jonas.Eschle@cern.ch>,"
            " Albert Puig <albert.puig.navarro@gmail.com>,"
            "Rafael Silva Coutinho <rsilvaco@cern.ch>,"
            "Matthieu Marinangeli <matthieu.marinangeli@cern.ch>"),
    install_requires=requirements,
    tests_require=tests_require,
    extras_require={
        'dev': requirements_dev + tests_require
    },
    use_scm_version=True,
)



"""Top-level package for zfit."""

#  Copyright (c) 2021 zfit
import inspect
import sys
import warnings

from pkg_resources import get_distribution

__version__ = get_distribution(__name__).version

__license__ = "BSD 3-Clause"
__copyright__ = "Copyright 2018, zfit"
__status__ = "Beta"

__author__ = ("Jonas Eschle <Jonas.Eschle@cern.ch>,"
              "Albert Puig <apuignav@gmail.com>, "
              "Rafael Silva Coutinho <rsilvaco@cern.ch>, "
              "Matthieu Marinangeli <matthieu.marinangeli@cern.ch>")
__maintainer__ = "zfit"
__email__ = 'zfit@physik.uzh.ch'
__credits__ = "Chris Burr, Martina Ferrillo, Abhijit Mathad, Oliver Lantwin, Johannes Lade"

__all__ = ["z", "constraint", "pdf", "minimize", "loss", "core", "data", "func", "dimension", "exception",
           "sample",
           "Parameter", "ComposedParameter", "ComplexParameter", "convert_to_parameter",
           "Space", "convert_to_space", "supports",
           "run", "settings"]

#  Copyright (c) 2019 zfit

if sys.version_info < (3, 7):
    msg = inspect.cleandoc(
        """zfit is being actively developed and keeps up with the newest versions of other packages.
        This includes Python itself. Therefore, Python 3.6 will be dropped in the near future (beginning of May 2021)
        and 3.9 will be added to the supported versions.

        Feel free to contact us in case of problems to upgrade to a more recent version of Python.
        """
    )
    warnings.warn(msg, FutureWarning, stacklevel=2)


def _maybe_disable_warnings():
    import os
    true = "IS_TRUE"
    if not os.environ.get("ZFIT_DISABLE_TF_WARNINGS", true):
        return
    elif true:
        warnings.warn("All TensorFlow warnings are by default suppressed by zfit."
                      " In order to not suppress them,"
                      " set the environment variable ZFIT_DISABLE_TF_WARNINGS to 0."
                      " In order to suppress the TensorFlow warnings AND this warning,"
                      " set ZFIT_DISABLE_TF_WARNINGS manually to 1.")
    os.environ["KMP_AFFINITY"] = "noverbose"
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

    import tensorflow as tf

    tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)


_maybe_disable_warnings()

import tensorflow as tf

if int(tf.__version__[0]) < 2:
    raise RuntimeError(f"You are using TensorFlow version {tf.__version__}. This zfit version ({__version__}) works"
                       f" only with TF >= 2")

from . import (constraint, core, data, dimension, exception, func, loss,
               minimize, param, pdf, sample, z)
from .core.data import Data
from .core.parameter import (ComplexParameter, ComposedParameter, Parameter,
                             convert_to_parameter)
from .core.space import Space, convert_to_space, supports
from .settings import run, ztypes
from .util.graph import jit as _jit


def _maybe_disable_jit():
    import os
    arg1 = os.environ.get("ZFIT_DO_JIT")
    arg2 = os.environ.get("ZFIT_EXPERIMENTAL_DO_JIT")
    arg3 = os.environ.get("ZFIT_MODE_GRAPH")
    if arg3 is not None:
        warnings.warn("Depreceated to use `ZFIT_MODE_GRAPH`, use `ZFIT_GRAPH_MODE` instead.",
                      DeprecationWarning)

    if arg1 is not None and arg2 is None:
        warnings.warn("Depreceated to use `ZFIT_EXPERIMENTAL_DO_JIT`, use `ZFIT_GRAPH_MODE` instead.",
                      DeprecationWarning)
    arg = arg2 if arg1 is None else arg1
    if arg is not None:
        run.set_graph_mode(bool(int(arg)))

    graph = os.environ.get("ZFIT_GRAPH_MODE")
    if graph is not None:
        run.set_graph_mode(bool(int(graph)))


# experimental flags


_maybe_disable_jit()

# EOF
